#include "hardware.h"

#define PAGE_SIZE 1056

unsigned char rwSPI0( unsigned char v )
{
	while ( ! ( AVR32_SPI0.sr & AVR32_SPI_TXEMPTY_MASK ) ) ;
	AVR32_SPI0.tdr = v;
	while ( ! ( AVR32_SPI0.sr & AVR32_SPI_RDRF_MASK ) ) ;
	return ( AVR32_SPI0.rdr & 0xFF );
}

void setupSerialFlash( void )
{
	/* setup SPI0 */	
	AVR32_SPI0.cr = AVR32_SPI_SWRST_MASK;
	
	AVR32_SPI0.csr0 = 
		( 2 << AVR32_SPI_SCBR ) |
		( AVR32_SPI_BITS_8_BPT << AVR32_SPI_BITS ) |
		AVR32_SPI_CSAAT_MASK |
		AVR32_SPI_NCPHA_MASK;
	
	AVR32_SPI0.mr = 
		AVR32_SPI_MSTR_MASK | 
		AVR32_SPI_MODFDIS_MASK | 
		( 10 << AVR32_SPI_DLYBCS );
	AVR32_SPI0.cr = AVR32_SPI_SPIEN_MASK;
}

void closeSerialFlash( void )
{	
	AVR32_SPI0.cr = AVR32_SPI_LASTXFER_MASK;
	udelay( 1 ); 
	AVR32_SPI0.cr = AVR32_SPI_SPIDIS_MASK;	
}

unsigned char readSerialFlashStatus( void )
{
	unsigned char res;
	setupSerialFlash();
	rwSPI0( 0xD7 );
	res = rwSPI0( 0 );
	closeSerialFlash();
	return res;
}

void issueFlashCommandAndWait( unsigned char opcode, int addr )
{	
	setupSerialFlash();
	rwSPI0( opcode );
	rwSPI0( ( addr >> 16 ) & 0xFF ); 
	rwSPI0( ( addr >> 8 ) & 0xFF ); 
	rwSPI0( ( addr >> 0 ) & 0xFF ); 
	closeSerialFlash();
	
	while ( ! ( readSerialFlashStatus() & 0x80 ) )
		udelay( 50 );	
}

void writeToSerialFlash( int offset, void *buffer, int size )
{
	int i;
	unsigned char *ptr = ( unsigned char * ) buffer;
	int page, byte, addr, first_page, last_page, pc;
		
	page = offset / PAGE_SIZE;
	byte = offset % PAGE_SIZE;
	pc = 0;

	/* determine first sector's start page */	
	if ( page < 8 ) {
		/* sector 0 */
		first_page = 0;
	}
	else if ( page < 256 ) {
		/* sector 1 */
		first_page = 8;
	}
	else {
		/* sector 2 - 32 */
		first_page = ( page & 0xFF );
	}
	
	/* determine last sector's end page */
	last_page = page + ( size - 1 ) / PAGE_SIZE;
	
	if ( last_page < 8 ) {
		/* sector 0 */
		last_page = 7;
	}
	else if ( last_page < 256 ) {
		/* sector 1 - 32 */
		last_page |= 0xFF;
	}
	
	/* to ensure data integrity we need to (re-program) the whole sector */
	/* see Atmel DataFlash Application note AN-4 */	
	for ( i = first_page; i <= last_page; i++ )
	{
		/* anything to write at all in this page? */
		if ( size  && i >= page ) {
			
			/* write partial page */
			if ( byte || size < PAGE_SIZE ) {
				/* first read page from flash into buffer 1 */
#if PAGE_SIZE == 1024
				issueFlashCommandAndWait( 0x53, i << 10 );
#else
				issueFlashCommandAndWait( 0x53, i << 11 );
#endif
			
				/* now change the bytes needed */
				addr = byte % PAGE_SIZE;
					
				setupSerialFlash();
				
				rwSPI0( 0x84 ); 
				rwSPI0( ( addr >> 16 ) & 0xFF ); 
				rwSPI0( ( addr >> 8 ) & 0xFF ); 
				rwSPI0( ( addr >> 0 ) & 0xFF );
					
				while ( byte < PAGE_SIZE && size ) {
					rwSPI0( *ptr++ );
					size--;
					byte++;
				}
					 
				closeSerialFlash();
					
				while ( ! ( readSerialFlashStatus() & 0x80 ) )
					udelay( 50 );	
						
				/* program buffer to flash */
#if PAGE_SIZE == 1024
				issueFlashCommandAndWait( 0x83, i << 10 );
#else
				issueFlashCommandAndWait( 0x83, i << 11 );
#endif
				byte = 0;
			} else {
				/* write whole page */

#if PAGE_SIZE == 1024
				addr = ( i << 10 );
#else
				addr = ( i << 11 );
#endif
					
				setupSerialFlash();
				rwSPI0( 0x82 ); 
				rwSPI0( ( addr >> 16 ) & 0xFF ); 
				rwSPI0( ( addr >> 8 ) & 0xFF ); 
				rwSPI0( ( addr >> 0 ) & 0xFF );
					
				while ( byte < PAGE_SIZE && size ) {
					rwSPI0( *ptr++ );
					size--;
					byte++;
				}
					 
				closeSerialFlash();
					
				while ( ! ( readSerialFlashStatus() & 0x80 ) )
					udelay( 50 );	
				
				byte = 0;
			}
		} else {	
			/* auto page rewrite  */
#if PAGE_SIZE == 1024
			issueFlashCommandAndWait( 0x58, i << 10 );
#else
			issueFlashCommandAndWait( 0x58, i << 11 );
#endif
		}	
	}
}

void readFromSerialFlash( int offset, void *buffer, int size )
{
	unsigned char *ptr = ( unsigned char * ) buffer;
#if PAGE_SIZE != 1024
	int page, byte;
	page = offset / PAGE_SIZE;
	byte = offset % PAGE_SIZE;
	offset = ( page << 11 ) | byte; 
#endif
	
	setupSerialFlash();

	rwSPI0( 0x0B );
	rwSPI0( ( offset >> 16 ) & 0xFF ); 
	rwSPI0( ( offset >> 8 ) & 0xFF ); 
	rwSPI0( ( offset >> 0 ) & 0xFF );
	rwSPI0( 0 );
	
	while ( size-- ) {
		*ptr++ = rwSPI0( 0 );
	}	
	
	closeSerialFlash();
}
