#include "hardware.h"
#include "lib.h"

/* Standard Request Codes */
#define GET_STATUS 					0
#define CLEAR_FEATURE 				1
#define SET_FEATURE 				2
#define SET_ADDRESS 				5
#define GET_DESCRIPTOR 				6
#define SET_DESCRIPTOR 				7
#define GET_CONFIGURATION 			8
#define SET_CONFIGURATION 			9
#define GET_INTERFACE 				10
#define SET_INTERFACE 				11
#define SYNCH_FRAME 				12

/* Descriptor Types */
#define DEVICE						1
#define CONFIGURATION				2
#define STRING						3
#define INTERFACE					4
#define ENDPOINT					5
#define DEVICE_QUALIFIER			6
#define OTHER_SPEED_CONFIGURATION	7
#define INTERFACE_POWER 			8

/* Standard Feature Selectors */
#define DEVICE_REMOTE_WAKEPUP		1
#define ENDPOINT_HALT				0
#define TEST_MODE					2

typedef struct {
	unsigned char bmRequestType;
	unsigned char bRequest;
	unsigned short wValue;
	unsigned short wIndex;
	unsigned short wLength;
} setupData;

const char myDevDesc[] = {
	0x12, 				/* bLength */
	DEVICE, 			/* bDescriptorType */
	0x00, 0x02, 		/* bcdUSB */
	0x00, 				/* bDeviceClass */
	0x00, 				/* bDeviceSubClass */
	0x00, 				/* bDeviceProtocol */
	0x40, 				/* bMaxPacketSize0 */
	0x66, 0x66,			/* idVendor */
	0x01, 0x00, 		/* idProduct */
	0x00, 0x01, 		/* bcdDevice */
	0x01, 				/* iManufacturer */
	0x02, 				/* iProduct */
	0x00, 				/* iSerialNumber */
	0x01				/* bNumConfigurations */
};

const char myConfDesc[] =  {
	0x09, 				/* bLength */
	CONFIGURATION, 		/* bDescriptorType */
	0x20, 0x00, 		/* wTotalLength */
	0x01, 				/* bNumInterfaces */
	0x01, 				/* bConfigurationValue */
	0x00, 				/* iConfiguration */
	0xC0, 				/* bmAttributes */
	0x00, 				/* bMaxPower */

	0x09,				/* bLength */ 
	INTERFACE, 			/* bDescriptorType */
	0x00, 				/* bInterfaceNumber */
	0x00, 				/* bAlternateSetting */
	0x02, 				/* bNumEndpoints */
	0xFF, 				/* bInterfaceClass */
	0xFF, 				/* bInterfaceSubClass */
	0xFF, 				/* bInterfaceProtocol */
	0x00,				/* iInterface */

	0x07, 				/* bLength */
	ENDPOINT, 			/* bDescriptorType */
	0x01, 				/* bEndpointAddress */
	0x02, 				/* bmAttributes */
	0x00, 0x02, 		/* wMaxPacketSize */
	0x0A,				/* bInterval */

	0x07, 				/* bLength */
	ENDPOINT, 			/* bDescriptorType */
	0x82, 				/* bEndpointAddress */
	0x02, 				/* bmAttributes */
	0x00, 0x02, 		/* wMaxPacketSize */
	0x0A				/* bInterval */
};

/***********
 * WARNING *
 ***********/

/* The max. transfer size for our control endpoint is 64 and
 * the current routine cannot transfer more than that.
 * 
 * If the strings below are too long, the routine needs to be
 * extended to split the transfers in 64 byte blocks */

const char string0 [] = { 
	4, 					/* bLength */
	STRING, 			/* bDescriptorType */
	0x07, 0x04			/* LANGID */
};

const char string1 [] = {
	32, 				/* bLength */
	STRING, 			/* bDescriptorType */
	'c', 0, 'o', 0, 'd', 0, 'e', 0, '-', 0, 
	'w', 0, 'i', 0, 'z', 0, 'a', 0, 'r', 0, 'd', 0, 
	'.', 0, 'c', 0, 'o', 0, 'm', 0
};

const char string2 [] = {
	30, 				/* bLength */
	STRING, 			/* bDescriptorType */
	'l', 0, 'o', 0, 'g', 0, 'i', 0, 'c', 0, ' ', 0, 
	'a', 0, 'n', 0, 'a', 0, 'l', 0, 'y', 0, 'z', 0, 'e', 0, 'r', 0 	
};

void initUSB( void )
{
	/* enable controller */
	AVR32_USBA.ctrl = AVR32_USBA_EN_USBA_MASK; 
	
	/* reset endpoints */
	AVR32_USBA.eptrst = 0xFF;
	
	/* enable interrupts */
	AVR32_USBA.ien = 
		AVR32_USBA_ENDRESET_MASK;
}

void controlEndpointInterrupt( void )
{
	int status;

	status = AVR32_USBA.endpoint[ 0 ].eptsta;
	
	if ( status & AVR32_USBA_RX_SETUP_MASK ) {
		setupData sdata;
		int byte_count;

		byte_count = 
			( status & AVR32_USBA_BYTE_COUNT_MASK ) >>
			AVR32_USBA_BYTE_COUNT;
			
		mymemcpy( &sdata, ( const void * )AVR32_USBA_FIFO, sizeof( sdata ) );
		sdata.wValue = ( sdata.wValue << 8 ) | ( sdata.wValue >> 8 );
		sdata.wIndex = ( sdata.wIndex << 8 ) | ( sdata.wIndex >> 8 );
		sdata.wLength = ( sdata.wLength << 8 ) | ( sdata.wLength >> 8 );		
		
		AVR32_USBA.endpoint[ 0 ].eptclrsta = AVR32_USBA_RX_SETUP_MASK;
		
		if ( sdata.bRequest == GET_DESCRIPTOR ) {
			if ( ( sdata.wValue >> 8 ) == DEVICE ) {

				while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;

				mymemcpy( ( void *) AVR32_USBA_FIFO, &myDevDesc, sizeof( myDevDesc ) );

				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;

			} else if ( ( sdata.wValue >> 8 ) == CONFIGURATION ) {

				while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;

				mymemcpy( ( void *) AVR32_USBA_FIFO, &myConfDesc, min( sdata.wLength, sizeof( myConfDesc ) ) );

				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;

			} else if ( ( sdata.wValue >> 8 ) == STRING ) {
				
				while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;

				switch ( sdata.wValue & 0xFF )
				{
				case 0:
					mymemcpy( ( void *) AVR32_USBA_FIFO, &string0, min( sdata.wLength, sizeof( string0 ) ) );
					break;
				case 1:
					mymemcpy( ( void *) AVR32_USBA_FIFO, &string1, min( sdata.wLength, sizeof( string1 ) ) );
					break;
				case 2:
					mymemcpy( ( void *) AVR32_USBA_FIFO, &string2, min( sdata.wLength, sizeof( string2 ) ) );
					break;
				}
				
				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;	
							
			} else {
				/* for all other descriptor types send back a stall */
				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_FRCESTALL_MASK;			

				while ( ! ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_STALL_SNT_MASK ) ) ;

				AVR32_USBA.endpoint[ 0 ].eptclrsta = AVR32_USBA_TOGGLESQ_MASK;				
			}
		}
		
		if ( sdata.bRequest == SET_ADDRESS ) {

			AVR32_USBA.ctrl |= sdata.wValue & 0x7F ; 

			while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;
			AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;
			while ( ! ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_COMPLT_MASK ) ) ;

			/* now could be a good time to set up our remaining endpoints */
			AVR32_USBA.endpoint[ 1 ].eptcfg = 
				( AVR32_USBA_ONE << AVR32_USBA_BK_NUMBER ) |
				( AVR32_USBA_EPT_TYPE_BULK << AVR32_USBA_EPT_TYPE ) |
				( AVR32_USBA_SZ_512 << AVR32_USBA_EPT_SIZE );
	
			AVR32_USBA.endpoint[ 2 ].eptcfg = 
				( AVR32_USBA_ONE << AVR32_USBA_BK_NUMBER ) |
				( AVR32_USBA_EPT_TYPE_BULK << AVR32_USBA_EPT_TYPE ) |
				AVR32_USBA_EPT_DIR_MASK |
				( AVR32_USBA_SZ_512 << AVR32_USBA_EPT_SIZE );
	
			AVR32_USBA.endpoint[ 1 ].eptctlenb = AVR32_USBA_EPT_ENABL_MASK | 
											     AVR32_USBA_RX_BK_RDY_MASK;
	
			AVR32_USBA.endpoint[ 2 ].eptctlenb = AVR32_USBA_EPT_ENABL_MASK;
	
			AVR32_USBA.ien |= AVR32_USBA_EPT_INT_1_MASK;

			AVR32_USBA.ctrl |= AVR32_USBA_FADDR_EN_MASK;			
		}
		
		if ( sdata.bRequest == SET_CONFIGURATION ) {
			while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;
			AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;

			/* reset endpoints */
			AVR32_USBA.eptrst = 0x6;
		}

		if ( sdata.bRequest == SET_INTERFACE ) { 
			AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_FRCESTALL_MASK;			

			while ( ! ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_STALL_SNT_MASK ) ) ;

			AVR32_USBA.endpoint[ 0 ].eptclrsta = AVR32_USBA_TOGGLESQ_MASK;				
		}
		
		if ( sdata.bRequest == CLEAR_FEATURE ) {
			if ( ( sdata.bmRequestType & 0x1F ) == 2 ) {
				int ep;
				/* a clear HALT is requested on an endpoint */
				if ( sdata.wIndex == 1 )
					ep = 1; 
				else 
					ep = 2;
					
				AVR32_USBA.endpoint[ ep ].eptclrsta = 
					AVR32_USBA_TOGGLESQ_MASK |
					AVR32_USBA_FRCESTALL_MASK;	

				while ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;
				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;
			} else {
				AVR32_USBA.endpoint[ 0 ].eptsetsta = AVR32_USBA_FRCESTALL_MASK;			
	
				while ( ! ( AVR32_USBA.endpoint[ 0 ].eptsta & AVR32_USBA_STALL_SNT_MASK ) ) ;
	
				AVR32_USBA.endpoint[ 0 ].eptclrsta = AVR32_USBA_TOGGLESQ_MASK;				
			}
		}
	}
}

void usbSend( const char *buf, int sz )
{
	while ( AVR32_USBA.endpoint[ 2 ].eptsta & AVR32_USBA_TX_PK_RDY_MASK ) ;
	mymemcpy( ( void * )&AVR32_USBA_FIFO[ 1 << 17 ], buf, sz );
	AVR32_USBA.endpoint[ 2 ].eptsetsta = AVR32_USBA_TX_PK_RDY_MASK;	
}

void bulkOutInterrupt( void )
{
	int status;
	char buf[ 512 ];

	status = AVR32_USBA.endpoint[ 1 ].eptsta;
	
	if ( status & AVR32_USBA_RX_BK_RDY_MASK ) {
		int byte_count;

		byte_count = 
			( status & AVR32_USBA_BYTE_COUNT_MASK ) >>
			AVR32_USBA_BYTE_COUNT;
		
		mymemcpy( buf, ( const void * )&AVR32_USBA_FIFO[ 1 << 16 ], byte_count );

		AVR32_USBA.endpoint[ 1 ].eptclrsta = AVR32_USBA_RX_BK_RDY_MASK;
		
		/* todo: process the data sent to us */
	}			
}

void usbaInterrupt( void )
{
	int globalStatus;
	
	globalStatus = AVR32_USBA.intsta;
	
	/* end of reset */
	if ( globalStatus & AVR32_USBA_ENDRESET_MASK ) {
		AVR32_USBA.clrint = AVR32_USBA_ENDRESET_MASK;
				
		/* setup control endpoint */
		AVR32_USBA.endpoint[ 0 ].eptcfg = 
			( AVR32_USBA_ONE << AVR32_USBA_BK_NUMBER ) |
			( AVR32_USBA_EPT_TYPE_CONTROL << AVR32_USBA_EPT_TYPE ) |
			( AVR32_USBA_SZ_64 << AVR32_USBA_EPT_SIZE );

		AVR32_USBA.endpoint[ 0 ].eptctlenb = AVR32_USBA_EPT_ENABL_MASK | AVR32_USBA_RX_SETUP_MASK;

		AVR32_USBA.ien |= AVR32_USBA_EPT_INT_0_MASK;
		
		/* enable controller */
		AVR32_USBA.ctrl = AVR32_USBA_EN_USBA_MASK; 
	}
	
	if ( globalStatus & AVR32_USBA_EPT_INT_0_MASK ) {
		AVR32_USBA.clrint = AVR32_USBA_EPT_INT_0_MASK;
		controlEndpointInterrupt();
	}	

	if ( globalStatus & AVR32_USBA_EPT_INT_1_MASK ) {
		AVR32_USBA.clrint = AVR32_USBA_EPT_INT_1_MASK;
		bulkOutInterrupt();
	}	
}

